"""
This module contains items related to data, such as raw csv's and data objects
"""

from . import load, configurable_dataset, dataset
from .adult import Adult
from .compas import Compas
from .credit import Credit
from .german import German
from .sqf import Sqf
from .toy import Toy
from .non_binary_toy import NonBinaryToy
