"""Pre-process algorithms take the training data and transform it"""
from .beutel import Beutel
from .pre_algorithm import PreAlgorithmAsync, PreAlgorithm
from .zemel import Zemel
from .vfae import VFAE
